#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cmath>
#include <ctime>
#include <CL/cl.hpp>
#include <petsc.h>
#include <petscviennacl.h>
#include "constants.h"

viennacl::ocl::program program;

extern void InitialCondition(TS ts, Vec prim);
extern PetscErrorCode ComputeResidualViennaCL(TS ts,
                               PetscScalar t,
                               Vec Prim, Vec dPrim_dt,
                               Vec F, void *ptr);
extern PetscErrorCode ComputeResidual(TS ts,
                               PetscScalar t,
                               Vec Prim, Vec dPrim_dt,
                               Vec F, void *ptr);

int main(int argc, char **argv)
{
    TS ts;
    Vec soln;
    DM da;

    static const char help[] = 
        "Petsc ViennaCL test";
    PetscInitialize(&argc, &argv, PETSC_NULL, help);

    DMDACreate2d(PETSC_COMM_WORLD, 
                 DMDA_BOUNDARY_GHOSTED, DMDA_BOUNDARY_GHOSTED,
                 DMDA_STENCIL_STAR,
                 N1, N2,
                 PETSC_DECIDE, PETSC_DECIDE,
                 1, NG, PETSC_NULL, PETSC_NULL, &da);

    DMDASetUniformCoordinates(da, 
                              X1_MIN + DX1/2., 
                              X1_MAX - DX1/2.,
                              X2_MIN + DX2/2.,
                              X2_MAX - DX2/2.,
                              0., 0.);

//    DMSetVecType(da, VECVIENNACL);
//    DMSetMatType(da, MATAIJVIENNACL);

    DMCreateGlobalVector(da, &soln);

    TSCreate(PETSC_COMM_WORLD, &ts);
    TSSetDM(ts, da);
    TSSetIFunction(ts, PETSC_NULL, ComputeResidual, NULL);
//    TSSetIFunction(ts, PETSC_NULL, ComputeResidualViennaCL, NULL);

    std::ifstream sourceFile("computeresidual.cl");
    std::string sourceCode((std::istreambuf_iterator<char>(sourceFile)),
                            std::istreambuf_iterator<char>());

    PetscScalar start = std::clock();
    program =
        viennacl::ocl::current_context().add_program(sourceCode,
                                                     "computeresidual");
    PetscScalar end = std::clock();
    PetscScalar time = (end - start)/(PetscScalar)CLOCKS_PER_SEC;
    PetscPrintf(PETSC_COMM_WORLD, 
                "Time taken for kernel compilation = %f\n", time);
    
    program.add_kernel("ComputeResidual");

    InitialCondition(ts, soln);

    TSSetSolution(ts, soln);
    TSSetFromOptions(ts);

    TSSolve(ts, soln);

    DMDestroy(&da);
    VecDestroy(&soln);
    TSDestroy(&ts);

    PetscFinalize();
    return(0);
}

PetscErrorCode ComputeResidualViennaCL(TS ts,
                                       PetscScalar t,
                                       Vec X, Vec dX_dt,
                                       Vec F, void *ptr)
{
    const viennacl::vector<PetscScalar> *x, *dx_dt;
    viennacl::vector<PetscScalar> *f;

    VecViennaCLGetArrayRead(X, &x);
    VecViennaCLGetArrayRead(dX_dt, &dx_dt);
    VecViennaCLGetArrayWrite(F, &f);

    viennacl::ocl::kernel kernel = program.get_kernel("ComputeResidual");

    kernel.global_work_size(0, N1);
    kernel.global_work_size(1, N2);

    viennacl::ocl::enqueue(kernel(*x, *dx_dt, *f));
    viennacl::backend::finish();

    VecViennaCLRestoreArrayRead(X, &x);
    VecViennaCLRestoreArrayRead(dX_dt, &dx_dt);
    VecViennaCLRestoreArrayWrite(F, &f);
}

PetscErrorCode ComputeResidual(TS ts,
                               PetscScalar t,
                               Vec X, Vec dX_dt,
                               Vec F, void *ptr)
{
    PetscScalar *x, *dx_dt, *f;
    VecGetArray(X, &x);
    VecGetArray(dX_dt, &dx_dt);
    VecGetArray(F, &f);
    
    for (int j=0; j<N2-1; j++)
        for (int i=0; i<N1-1; i++) {
            f[INDEX_GLOBAL(i,j)] = dx_dt[INDEX_GLOBAL(i,j)] -
                                     (x[INDEX_GLOBAL(i+1,j)] - 
                                      x[INDEX_GLOBAL(i,j)])/DX1 -
                                     (x[INDEX_GLOBAL(i,j+1)] - 
                                      x[INDEX_GLOBAL(i,j)])/DX2;
        }
    VecRestoreArray(X, &x);
    VecRestoreArray(dX_dt, &dx_dt);
    VecRestoreArray(F, &f);
}

void InitialCondition(TS ts, Vec X)
{
    DM da;
    PetscScalar *x;

    int X1Start, X2Start;
    int X1Size, X2Size;
    int X1End, X2End;

    TSGetDM(ts, &da);

    VecGetArray(X, &x);

    DMDAGetCorners(da, 
                   &X1Start, &X2Start, NULL,
                   &X1Size, &X2Size, NULL);

    X1End = X1Start + X1Size;
    X2End = X2Start + X2Size;

    for (PetscInt j=X2Start; j<X2End; j++) {
        for (PetscInt i=X1Start; i<X1End; i++) {
            
            PetscScalar X1Coord = X1_MIN + DX1/2. + i*DX1;
            PetscScalar X2Coord = X2_MIN + DX2/2. + j*DX2;

            PetscScalar X1Center = (X1_MIN + X1_MAX)/2.;
            PetscScalar X2Center = (X2_MIN + X2_MAX)/2.;

            PetscScalar r = PetscSqrtScalar(
                                PetscPowScalar(X1Coord-X1Center, 2.0) + 
                                PetscPowScalar(X2Coord-X2Center, 2.0));

            x[i + j*N1] = exp(-r*r/.01);

        }
    }

    VecRestoreArray(X, &x);
}

